/*
  PIM for Quagga
  Copyright (C) 2008  Everton da Silva Marques

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; see the file COPYING; if not, write to the
  Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
  MA 02110-1301 USA
  
  $Id$
*/

#include <zebra.h>
#include "zebra/rib.h"

#include "log.h"
#include "prefix.h"
#include "zclient.h"
#include "stream.h"
#include "network.h"
#include "thread.h"

#include "pimd.h"
#include "pim_str.h"
#include "pim_zlookup.h"

/* Connect to zebra for nexthop lookup. */
static int zclient_lookup_connect(struct thread *t)
{
  struct zclient *zlookup;

  zlookup = THREAD_ARG(t);
  zlookup->t_connect = NULL;

  if (zlookup->sock >= 0) {
    return 0;
  }

#ifdef HAVE_TCP_ZEBRA
  zlookup->sock = zclient_socket();
  if (zlookup->sock < 0) {
    zlog_warn("%s: failure connecting TCP socket %s,%d",
	      __PRETTY_FUNCTION__, "127.0.0.1", ZEBRA_PORT);
    return -1;
  }
  zlog_notice("%s: connected TCP socket %s,%d",
	      __PRETTY_FUNCTION__, "127.0.0.1", ZEBRA_PORT);
#else
  zlookup->sock = zclient_socket_un(ZEBRA_SERV_PATH);
  if (zlookup->sock < 0) {
    zlog_warn("%s: failure connecting UNIX socket %s",
	      __PRETTY_FUNCTION__, ZEBRA_SERV_PATH);
    return -1;
  }
  zlog_notice("%s: connected UNIX socket %s",
	      __PRETTY_FUNCTION__, ZEBRA_SERV_PATH);
#endif /* HAVE_TCP_ZEBRA */

  return 0;
}

/* Schedule connection. */
static void zclient_lookup_sched(struct zclient *zlookup)
{
  assert(!zlookup->t_connect);
  zlookup->t_connect = thread_add_event(master,
					zclient_lookup_connect,
					zlookup, 0);
  zlog_notice("%s: zclient lookup connection scheduled",
	      __PRETTY_FUNCTION__);
}

/* Schedule reconnection, if needed. */
static void zclient_lookup_reconnect(struct zclient *zlookup)
{
  if (zlookup->t_connect) {
    return;
  }

  zclient_lookup_sched(zlookup);
}

struct zclient *zclient_lookup_new()
{
  struct zclient *zlookup;

  zlookup = zclient_new();
  if (!zlookup) {
    zlog_err("%s: zclient_new() failure",
	     __PRETTY_FUNCTION__);
    return 0;
  }

  zlookup = zclient_new();
  zlookup->sock = -1;
  zlookup->ibuf = stream_new(ZEBRA_MAX_PACKET_SIZ);
  zlookup->obuf = stream_new(ZEBRA_MAX_PACKET_SIZ);
  zlookup->t_connect = 0;

  zclient_lookup_sched(zlookup);

  zlog_notice("%s: zclient lookup socket initialized",
	      __PRETTY_FUNCTION__);

  return zlookup;
}

static int zclient_read_nexthop(struct zclient *zlookup, int ifindex[],
				const int ifindex_size, struct in_addr addr)
{
  int num_ifindex = 0;
  struct stream *s;
  uint16_t length;
  u_char marker;
  u_char version;
  uint16_t command;
  int nbytes;
  struct in_addr raddr;
  uint32_t metric;
  int i;
  u_char nexthop_num;

  s = zlookup->ibuf;
  stream_reset(s);

  nbytes = stream_read(s, zlookup->sock, 2);
  if (nbytes < 1) {
    zlog_err("%s %s: failure reading zclient lookup socket",
	     __FILE__, __PRETTY_FUNCTION__);
    close(zlookup->sock);
    zlookup->sock = -1;
    zclient_lookup_reconnect(zlookup);
    return -1;
  }
  length = stream_getw(s);

  nbytes = stream_read(s, zlookup->sock, length - 2);
  if (nbytes < 1) {
    zlog_err("%s %s: failure reading zclient lookup socket",
	     __FILE__, __PRETTY_FUNCTION__);
    close(zlookup->sock);
    zlookup->sock = -1;
    zclient_lookup_reconnect(zlookup);
    return -2;
  }
  marker = stream_getc(s);
  version = stream_getc(s);
  
  if (version != ZSERV_VERSION || marker != ZEBRA_HEADER_MARKER) {
    zlog_err("%s: socket %d version mismatch, marker %d, version %d",
	     __func__, zlookup->sock, marker, version);
    return -3;
  }
    
  command = stream_getw(s);
  
  raddr.s_addr = stream_get_ipv4(s);
  metric = stream_getl(s);
  nexthop_num = stream_getc(s);

  if (!nexthop_num) {
    return -4;
  }

  for (i = 0; i < nexthop_num; ++i) {
    enum nexthop_types_t nexthop_type = stream_getc(s);
    switch (nexthop_type) {
    case ZEBRA_NEXTHOP_IFINDEX:
    case ZEBRA_NEXTHOP_IFNAME:
    case ZEBRA_NEXTHOP_IPV4_IFINDEX:
      if (num_ifindex >= ifindex_size) {
	char addr_str[100];
	pim_inet4_dump("<addr?>", addr, addr_str, sizeof(addr_str));
	zlog_warn("%s %s: found too many nexthop ifindexes (%d > %d) for address %s",
		 __FILE__, __PRETTY_FUNCTION__,
		 (num_ifindex + 1), ifindex_size, addr_str);
	return num_ifindex;
      }
      if (nexthop_type == ZEBRA_NEXTHOP_IPV4_IFINDEX) {
	struct in_addr nexthop_addr;
	nexthop_addr.s_addr = stream_get_ipv4(s);
      }
      ifindex[num_ifindex++] = stream_getl(s);
      break;
    default:
      /* do nothing */
      {
	char addr_str[100];
	pim_inet4_dump("<addr?>", addr, addr_str, sizeof(addr_str));
	zlog_warn("%s %s: found non-ifindex nexthop type=%d for address %s",
		 __FILE__, __PRETTY_FUNCTION__,
		  nexthop_type, addr_str);
      }
      break;
    }
  }

  return num_ifindex;
}

int zclient_lookup_nexthop(struct zclient *zlookup, int ifindex[],
			   const int ifindex_size, struct in_addr addr)
{
  struct stream *s;
  int ret;

  /* Check socket. */
  if (zlookup->sock < 0) {
    zlog_err("%s %s: zclient lookup socket is not connected",
	     __FILE__, __PRETTY_FUNCTION__);
    zclient_lookup_reconnect(zlookup);
    return -1;
  }
  
  s = zlookup->obuf;
  stream_reset(s);
  zclient_create_header(s, ZEBRA_IPV4_NEXTHOP_LOOKUP);
  stream_put_in_addr(s, &addr);
  stream_putw_at(s, 0, stream_get_endp(s));
  
  ret = writen(zlookup->sock, s->data, stream_get_endp(s));
  if (ret < 0) {
    zlog_err("%s %s: writen() failure writing to zclient lookup socket",
	     __FILE__, __PRETTY_FUNCTION__);
    close(zlookup->sock);
    zlookup->sock = -1;
    zclient_lookup_reconnect(zlookup);
    return -2;
  }
  if (ret == 0) {
    zlog_err("%s %s: connection closed on zclient lookup socket",
	     __FILE__, __PRETTY_FUNCTION__);
    close(zlookup->sock);
    zlookup->sock = -1;
    zclient_lookup_reconnect(zlookup);
    return -3;
  }
  
  return zclient_read_nexthop(zlookup, ifindex, ifindex_size, addr);
}
